/*
 * packetmodeller/packet_modeller.c
 *
 * Copyright (C) 2008-2009 Reto Gantenbein
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

#include <errno.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <arpa/inet.h>
#include <net/if.h>
#include <netinet/in.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sys/types.h>
#include "logger.h"
#include "message.h"
#include "netdevice.h"
#include "packet_modeller.h"
#include "typedef.h"
#include "init.h"

/** Print usage details */
void
print_usage()
{
    printf("\nUsage:\t%s MODELIP:MODELPORT LOCALPORT\n", exec_name);
    printf("\t\t[-d][-h][-n][-p INTERFACE][-w INTERFACE[,INTERFACE]]\n");
    printf("Try `%s -h' for more details\n", exec_name);
    exit(1);
}

/** Print help */
void
print_help()
{
    printf("\nUsage:\t%s MODELIP:MODELPORT LOCALPORT\n", exec_name);
    printf("\t\t[-h][-d][-p INTERFACE][-w INTERFACE[,INTERFACE]]\n");
    printf("\nTunnels traffic from a local network to another %s", exec_name);
    printf("instance\nor a wireless emulation model\n");
    printf("\nParameters:\n");
    printf("MODELIP:PORT\tIP Address and port number of tunnel end point\n");
    printf("LOCALPORT\tLocal port to listen for connections\n");
    printf("-d\t\tEnable debug output\n");
    printf("-h\t\tShow this help message\n");
    printf("-n\t\tEnable node-to-node communication (only for testing!)\n");
    printf("-p INTERFACE\tTunnel interface (default: %s)\n", interfaces[0]);
    printf("-w INTERFACE\tLocal network interface for redirecting\n");
    printf("\t\tMust be a TUN interface (default: %s)\n", interfaces[1]);
    printf("\t\tUp to two interfaces are allowed: E.g. -w %s,%s\n", \
        interfaces[1], interfaces[2]);
    exit(0);
}

/** Set wireless information to Ethernet frames and send to model */
void
process_data_from_tap(struct node_msg *message, const int ether_length)
{
    (*message).header.type = NODE_MSG;
    memmove(&(*message).header.id, node_id, sizeof(host_id));
    (*message).channel = 1;  // TODO: set channel and transmit power dynamically
    (*message).transmit_power = 34;
    (*message).pkt_length = ether_length;
    dbg_msg((char *) message, ether_length);

    /* send data to model */
    if (sendto(net_socket, message, NODE_PREAMBLE_SIZE + ether_length, 0, \
        (struct sockaddr *) &model_addr, sizeof(struct sockaddr_in)) < 0)
    {
        printf("process_data_from_tap(): Sending data failed (%s)\n", \
            strerror(errno));
    }
    print_dbg("[net_socket] %i Bytes sent (%s)\n", \
        NODE_PREAMBLE_SIZE + ether_length, strerror(errno));
}

/** Process packets received from WlanModel and inject to local network */
void
process_data_from_net(char *buf, const int message_length)
{
    int payload_length;
    RAWPACKET *ether_pkt;

    // received message from model
    if (isModelMessage((struct msg_header *) buf))
    {
        struct model_msg *message = (struct model_msg *) buf;
        ether_pkt = (RAWPACKET *) &((*message).wlan_pkt);
        payload_length = message_length - MODEL_PREAMBLE_SIZE;
        dbg_msg(buf, payload_length);
    } else {
        // or other node
        if (isNodeMessage((struct msg_header *) buf))
        {
            struct node_msg *message = (struct node_msg *) buf;
            ether_pkt = (RAWPACKET *) &((*message).wlan_pkt);
            payload_length = message_length - NODE_PREAMBLE_SIZE;
            dbg_msg(buf, payload_length);
        } else {
            printf("process_data_from_net(): Received unknown message\n");
            return;
        }
    }

    // send to TAP interface
    if (write(tap_fd, ether_pkt, payload_length) < 0)
    {
        print_dbg("%s:%i - Couldn't locally deliver packet (%s)\n", \
            __FILE__, __LINE__, strerror(errno));
    }
    print_dbg("[tap_fd] %i Bytes sent (%s)\n", payload_length, strerror(errno));
}

/** Listen to TUN and network interfaces for received packets */
int
main_loop()
{
    fd_set fds;
    int recv_bytes;
    register int ret;
    // node message is biggest message possible
    size_t max_pkt_size = NODE_PREAMBLE_SIZE + MAX_MSG_SIZE;
    char buffer[max_pkt_size];
    fm = max(tap_fd, net_socket) + 1;

    for (;;)
    {
        FD_ZERO(&fds);
        FD_SET(tap_fd, &fds);
        FD_SET(net_socket, &fds);
        if ((ret = select(fm, &fds, NULL, NULL, NULL)) < 0)
        {
            print_dbg("%s:%i - Error in selecting file descriptor (%s)\n", \
                __FILE__, __LINE__, strerror(errno));
        } else {
            // check tap device for incoming traffic
            if (FD_ISSET(tap_fd, &fds))
            {
                struct node_msg *message = (struct node_msg *) buffer;
                recv_bytes = read(tap_fd, &(*message).wlan_pkt, MAX_MSG_SIZE);
                print_dbg("[tap_fd] %i Bytes read (%s)\n", \
                    recv_bytes, strerror(errno));
                if (recv_bytes < 0)
                {
                    printf("main_loop(): Error reading data from tap device ");
                    printf("(%s)\n", strerror(errno));
                } else {
                    process_data_from_tap(message, recv_bytes);
                }
            }
            // check network device for incoming traffic
            if (FD_ISSET(net_socket, &fds))
            {
                recv_bytes = recvfrom(net_socket, buffer, max_pkt_size, 0, \
                    NULL, NULL);
                print_dbg("[net_socket] %i Bytes read (%s)\n", \
                    recv_bytes, strerror(errno));
                if (recv_bytes < 0)
                {
                    printf("main_loop(): Error reading data from network ");
                    printf("device (%s)\n", strerror(errno));
                } else {
                    process_data_from_net(buffer, recv_bytes);
                }
            }
        }
    }
    return 0;
}

/** PacketModeller */
int
packet_modeller()
{
    init_fds();
    register_node();

    main_loop();

    return 0;
}

int
main(int argc, char *argv[])
{
    int err = 0;
    exec_name = argv[0];
    pars_opts(argc, argv);
    if ((err = init_node()) == 0)
    {
        err = packet_modeller();
    } else {
        printf("%s:%i - Error in initializing node\n", __FILE__, __LINE__);
    }
    exit(err);
}
