/*
 * wlanmodel/nodeinfo.cc
 *
 * Copyright (C) 2008-2009 Reto Gantenbein
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

#include <cstdlib>
#include <string>
#include <arpa/inet.h>
#include <netinet/in.h>
#include <sys/socket.h>
#include "nodeinfo.h"
#include "nodemanager.h"
#include "typedef.h"

/**
 * Constructors
 */
NodeInfo::NodeInfo() : cObject()
{
    setName("nodeinfo");
    hostId = "";
    hostName = "";
    virtMacAddr = "000000000000";
    hostPort = 0;
    memset(&hostIp, 0, sizeof(in_addr));
    proxyHostGate = -1;
}

NodeInfo::NodeInfo(const host_id *id, const std::string name, \
                    const in_addr *addr, const uint16_t port) : cObject()
{
    // set name as index for attaching NodeInfo to a message
    setName("nodeinfo");
    hostId.append((const char *) id);
    hostName = name;
    hostIp = *addr;
    hostPort = port;

    // check if there is already gate assigned to this hostId
    if ((proxyHostGate = NodeManager::getProxyGate(id)) == -1)
    {
        proxyHostGate = gateIndex;
        gateIndex++;
    }

    ev << str() << "\n";
}

/**
 * Destructor
 */
NodeInfo::~NodeInfo()
{
}

/**
 * Set initial gate index
 */
int NodeInfo::gateIndex = 0;

/**
 * Return 6 byte char representation of nodeId
 */
std::string
NodeInfo::getHostId()
{
    return hostId;
}

/**
 * Return host name
 */
std::string
NodeInfo::getHostName()
{
    return hostName;
}

/**
 * Return proxy gate number
 */
int
NodeInfo::getProxyGate()
{
    return proxyHostGate;
}

/**
 * Converts 6 byte char representation of nodeId to string
 */
void
NodeInfo::convertToHostIdString(const host_id *hostId, std::string *hostIdString)
{
    char hexId[12];
    int i;

    const char *idp = (char *) hostId;
    char *hexp = hexId;
    for (i = 0; i < 6; ++i) {
        sprintf(hexp, "%02hhx", idp[i]);
        hexp++;
        hexp++;
    }
    (*hostIdString) = std::string(hexId, 12);
}

/**
 * Fill sockaddr_in struct with destination information of this node
 */
void
NodeInfo::getNodeAddr(sockaddr_in *node_addr)
{
    memcpy(&(*node_addr).sin_addr, (char *) &hostIp, sizeof(in_addr));
    uint16_t n_port = hostPort;
    memcpy(&(*node_addr).sin_port, &n_port , sizeof(uint16_t));
}

/**
 * Return string-representation of NodeInfo object
 */
std::string
NodeInfo::str()
{
    std::ostringstream port, index;

    std::string nodeInfoString("NodeInfo(): ");
    nodeInfoString.append(hostName);
    nodeInfoString.append(" id=");
    nodeInfoString.append(hostId);
    nodeInfoString.append(", IP=");
    nodeInfoString.append(inet_ntoa(hostIp));
    nodeInfoString.append(", port=");
    port << ntohs(hostPort);
    nodeInfoString.append(port.str());
    nodeInfoString.append(", proxyHostGate=");
    index << proxyHostGate;
    nodeInfoString.append(index.str());
    return nodeInfoString;
}
